/*************************************************************************/
/*                                                                       */
/*	RAND.C y STAT.C                                                      */
/*	Random Variable Generation                                           */
/*  Versin 1.1                                                          */
/*************************************************************************/
/**** he puesto define CPU 8086 como comentario Pablo BORONAT****/

#include <math.h>

/*#define CPU 8086*/
#define CPU 68000
typedef double real;
#define then

#define A 16807L		/*multiplier (7**5) for 'ranf' */
#define M 2147483647L		/*modulus (2**31-1) for 'ranf' */

static long In[16]= {0L,	/* seeds for streams 1 thru 15 */
  1973272912L,  747177549L,   20464843L,  640830765L, 1098742207L,
    78126602L,   84743774L,  831312807L,  124667236L, 1172177002L,
  1124933064L, 1223960546L, 1878892440L, 1449793615L,  553303732L};

static int strm=-1;		/* index of current stream */

#if CPU==8086
/*-------------- UNIFORM [0,1] RANDOM NUMBER GENERATOR------------------*/
/*                                                                      */
/* this implementation is for the Intel 8086/8 and 80286/386 CPUs using */
/* C compilers with 16-bit short integers and 32 bit long integers.     */
/*                                                                      */
/*----------------------------------------------------------------------*/
real ranf()
  {
   short *p,*q,k; long Hi,Lo;
   /* generate product using double precision simulation (comments   */
   /* refer to In's lower 16-bits as "L", its upper 16 bits as "H")  */
   p=(short *)&In[strm]; Hi=*(p+1)*A;                 /* 16807*H->Hi */
   *(p+1)=0; Lo=In[strm]*A;			      /* 16807*L->Lo */
   p=(short *)&Lo; Hi+=*(p+1);	 /* add high-order bits of Lo to Hi  */
   q=(short *)&Hi;		        /* low order bits of Hi-> LO */
   *(p+1)=*q&0X7FFF;				/* clear sign bit    */
   k=*(q+1)<<1; if (*q&0X8000) then k++;        /* Hi bits 31-45->k  */
   /* form Z + K [-M] (where Z=Lo) : presubtract M to avoid overflow */
   Lo-=M; Lo+=k; if (Lo<0) then Lo+=M;
   In[strm]=Lo;
   return((real)Lo*4.656612875E-10);		/* Lo x 1/(2**31-1)  */
  }
#endif


#if CPU==68000
/*-------------UNIFORM [0,1] RANDOM NUMBER GENERATOR--------------------*/
/*                                                                      */
/*  this implementation is for motorola 680x0 CPUs using C compilers    */
/*  with 16-bit short integers and 32 bit long integers.                */
/************************************************************************/
real ranf()
  {
   short *p,*q,k; long Hi,Lo;
   /* generate product using double precision simulation ( comments  */
   /* refer to In's lower 16 bits as "L", its upper 16 bits as "H")  */
   p=(short *)&In[strm]; Hi=*(p)*A;		       /* 16807*H->Hi*/
   *(p)=0; Lo=In[strm]*A;			       /* 16807*L->Lo*/
   p=(short *)&Lo; Hi+=*(p);	  /* add high-order bits of Lo to HI */
   q=(short *)&Hi;			/* low-order bits of Hi-> Lo */
   *(p)=*(q+1)&0X7FFF;			         /* clear sign bit   */
   k=*(q)<<1; if (*(q+1)&0X8000) then k++;     /* Hi bits 31-45->k   */
   /* form Z + K [-M] (where Z=Lo): presubtract M to avoid overflow  */
   Lo-=M; Lo+=k; if (Lo<0) then  Lo+=M;
   In[strm]=Lo;
   return((real)Lo*4.656612875E-10);		/* Lo x 1/(2**31-1)  */
  }
#endif


/*-------------------SELECT GENERATOR STREAM------------------------------*/
stream(n)
  int n;
    { /* set stream for 1<=n<=15, return stream for n=0 */
     if ((n<0) || (n>15)) then error(0,"Stream Argument Error");
     if (n) then strm=n;
     return(strm);
    }


/*------------------------SET/GET SEED------------------------------------*/
long seed(Ik,n)
  long Ik; int n;
    { /* set seed of stream n for Ik>0, return current seed for Ik=0  */
      if ((n<1) || (n>15)) then error(0,"Seed Argument Error");
      if (Ik>0L) then In[n]=Ik;
      return(In[n]);
    }


/*--------------UNIFORM[a,b] RANDOM VARIABLE GENERATOR--------------------*/
real uniform(a,b)
  real a,b;
    { /* 'uniform'returns a psuedo-random variate from a uniform          */
      /* distribution with a lower bound a and upper bound b.		  */
      if (a>b) then error(0,"uniform Argument Error: a > b");
      return(a+(b-a)*ranf());
    }


/*--------------RANDOM INTEGER GENERATOR----------------------------------*/
random(i,n)
  int i,n;
    { /* 'random'returns an integer equiprobably selected from the    */
      /* set of integers i, i+1, i+2,...,n.			      */
      if (i>n) then error(0,"random argument error: i > n");
      n-=i; n=(n+1.0)*ranf();
      return(i+n);
    }


/*------------EXPONENTIAL RANDOM VARIATE GENERATOR------------------------*/
real expntl(x)
  real x;
    { /* 'expntl'returns a psuedo-random variate from a negative     */
      /* exponential distribution with mean x.			     */
      return(-x*log(ranf()));
    }


/*--------------ERLANG  RANDOM VARIATE GENERATOR--------------------------*/
real erlang(x,s)
  real x,s;
    { /* 'erlang'returns a psuedo-random variate from an erlang	*/
      /* distribution with mean x and standard desviation s	*/
      int i,k; real z;
      if (s>x) then error(0,"erlang argument error: s > x");
      z=x/s; k=(int)z*z;
      z=1.0; for (i=0; i<k; i++) z*=ranf();
      return(-(x/k)*log(z));
    }


/*------HYPEREXPONENTIAL RANDOM VARIATE GENERATION----------------------*/
real hiperx(x,s)
  real x,s;
    {  /* 'hiperex'returns a psuedo-random variate from Morse's rwo-stage*/
       /* hiperexponential distribution with mean x and standard desvia- */
       /* tion s, s>x.    						 */
       real cv,z,p;
       if (s<=x) then error(0,"hiperex Argument Error: s not > x");
       cv=s/x;  z=cv*cv; p=0.5*(1.0-sqrt((z-1.0)/(z+1.0)));
       z=(ranf()>p)? (x/(1.0-p)):(x/p);
       return(-0.5*z*log(ranf()));
    }


/*------------NORMAL RANDOM VARIATE GENERATOR----------------------------*/
real normal(x,s)
  real x,s;
    { /* 'normal'returns a psuedo-random variate from a normal distribu- */
      /* tion with a mean x and a desviation s.				 */
      real v1,v2,w,z1; static real z2=0.0;
      if (z2!=0.0)
	then {z1=z2; z2=0.0;}   /* use value from previous call */
	else
	  {
	    do
	      {v1=2.0*ranf()-1.0; v2=2.0*ranf()-1.0; w=v1*v1+v2*v2;}
	    while (w>=1.0);
	    w=sqrt((-2.0*log(w))/w); z1=v1*w; z2=v2*w;
	  }
      return(x+z1*s);
    }


/***********************************************************************/
/*                                                                     */
/*			File "STAT.C"                                              */
/*	Normal and T Distribution Quantile Computation Functions           */
/*                                                                     */
/***********************************************************************/
/*-------------- COMPUTE pth QUANTILE OF THE NORMAL DISTRIBUTION ------*/
real Z(p)
  real p;
    { /* This function computes the pth upper quantile of the standard */
      /* normal distribution (i.e., the value of z wich the are under  */
      /* the curve from z to +infinity is equal to p).  'Z'is a trans- */
      /* literation of the 'STDZ' function in Appendix C of "Principles*/
      /* of Discrete Event Simulation", G.S.Fishman, Wiley,1978. The   */
      /* approximation used initially appeared in "Approximations for  */
      /* digital computers", C. Hastings, Jr., Princeton U. Press,1955.*/
	  real q,z1,n,d; double sqrt(),log();
	  q=(p>0.5)? (1-p):p; z1=sqrt(-2.0*log(q));
      n=(0.010328*z1+0.802853)*z1+2.515517;
      d=((0.001308*z1+0.189269)*z1+1.43278)*z1+1;
	  z1-=n/d; if (p>0.5) then z1=-z1;
      return(z1);
    }


/*-------------- COMPUTE pth QUANTILE OF THE t DISTRIBUTION ------------*/
real T(p,ndf)
  real p; int ndf;
    { /* This function computes the upper pth quantile of the t distri- */
      /* bution (the value of t for wich the area under the curve from t*/
      /* to +infinity is equal to p). It is a transliteration of the    */
      /* 'STUDTP' function given in Appendix C of "Principles of Discre-*/
      /* te Event Simulation", G. S. Fishman, Wiley, 1978.		*/
      int i; real z1,z2,h[4],x=0.0; double fabs();
      z1=fabs(Z(p)); z2=z1*z1;
      h[0]=0.25*z1*(z2+1.0); h[1]=0.010416667*z1*((5.0*z2+16.0)*z2+3.0);
      h[2]=0.002604167*z1*(((3.0*z2+19.0)*z2+17.0)*z2-15.0);
      h[3]=z1*((((79.0*z2+776.0)*z2-1482.0)*z2-1920.0)*z2-945.0);
      h[3]*=0.000010851;
      for (i=3; i>=0; i--) x=(x+h[i])/(real)ndf;
      z1+=x; if (p>0.5) then z1=-z1;
      return(z1);
    }
